from datetime import datetime


def calculate_nb_of_blocks_until_date(future_date: str or datetime):
    if isinstance(future_date, str):
        future_date = datetime.strptime(future_date, '%Y-%m-%d %H:%M:%S')
    delay = future_date - datetime.now()
    nb_blocks = int(delay.seconds / 4.5)
    return nb_blocks



from typing import Union
from pathlib import Path
from algosdk.v2client import algod


class Connector:
    def __init__(
        self,
        network: str = 'testnet',
        credentials: Union[Path, str] = None,
        token: str = None
    ):
        self.network = network
        self.credentials = credentials
        self.token = token or self._get_token_from_credentials_file()
        self.headers = {'X-Api-key': self.token}

    def _get_token_from_credentials_file(self):
        """Credentials file is a 1-line file with only the token
        """
        with open(self.credentials) as h:
            return h.readline().strip()

    def connect(self):
        self.algod_client = algod.AlgodClient(self.token, self.algod_address, headers=self.headers)


class SandboxConnector(Connector):
    def __init__(
        self,
        token: str = 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
        url: str = 'localhost',
        port: int = 4001
    ):
        super().__init__(network='sandbox', token=token)
        self.algod_address = f'http://{url}:{port}'


class PureStakeConnector(Connector):
    def __init__(
        self,
        network: str = 'testnet',
        credentials: str = '../tmp/.purestake',
        algod_version: str = 'v2'
    ):
        super().__init__(network=network, credentials=credentials)
        self.algod_version = algod_version
        self.algod_address = f'https://{self.network}-algorand.api.purestake.io/ps2'



from abc import abstractmethod
from typing import Optional
from datetime import datetime
from algosdk import transaction, template
from algosdk.v2client.algod import AlgodClient
from algosdk.transaction import AssetConfigTxn, AssetTransferTxn


class Transactor:
    _cached_params = {}

    def __init__(
        self,
        client: str,
        sender_address: str,
        sender_private_key: str,
        fee: float = None,
        first_valid_round: int = None,
        last_valid_round: int = None,
        gh: bytes = None
    ):
        self.client = client
        self.sender_address = sender_address
        self.sender_private_key = sender_private_key
        self.fee = fee or self._get_fee()
        self.first_valid_round = first_valid_round or self._get_first_valid_round()
        self.last_valid_round = last_valid_round or self._get_last_valid_round()
        self.gh = gh or self._get_gh()

    def _params(self):
        self._cached_params = self.client.suggested_params()
        return self._cached_params

    @property
    def params(self):
        if not self._cached_params:
            self._params()
        return self._cached_params

    @abstractmethod
    def transact(self):
        pass

    def _get_fee(self):
        return self.params.fee

    def _get_first_valid_round(self):
        return self.params.first

    def _get_last_valid_round(self):
        return self.params.last

    def _get_gh(self):
        return self.params.gh

    def _wait_for_confirmation(self):
        """
        Utility function to wait until the transaction is
        confirmed before proceeding.
        """
        client = self.client
        txid = self.txid
        last_round = client.status().get('last-round')
        txinfo = client.pending_transaction_info(txid)
        while not (txinfo.get('confirmed-round') and txinfo.get('confirmed-round') > 0):
            print("Waiting for confirmation")
            last_round += 1
            client.status_after_block(last_round)
            txinfo = client.pending_transaction_info(txid)
        print("Transaction {} confirmed in round {}.".format(txid, txinfo.get('confirmed-round')))
        self.txinfo = txinfo
        return txinfo
        # last_round = algodclient.status().get('last-round')
        # while True:
        #    txinfo = algodclient.pending_transaction_info(txid)
        #    self.txinfo = txinfo
        #    self.asset_id = txinfo.get('asset-index')
        #    if not txinfo.get('round'):
        #        break
        #    if txinfo.get('round') and txinfo.get('round') > 0:
        #       print("Transaction {} confirmed in round {}.".format(txid, txinfo.get('round')))
        #       break
        #    else:
        #       print("Waiting for confirmation...")
        #       print('Round:', txinfo.get('round'))
        #       last_round += 1
        #       algodclient.status_after_block(last_round)
        # self.txinfo
        # return txinfo

    def sign_transaction(self, tx):
        signed_tx = tx.sign(self.sender_private_key)
        self.txid = signed_tx.transaction.get_txid()
        return signed_tx

    def send_transaction(self, signed_tx):
        try:
            tx_confirm = self.client.send_transaction(signed_tx)
            self._wait_for_confirmation()
            return tx_confirm
        except Exception as e:
            print(e)


class AssetOptinTransactor(Transactor):

    def __init__(
        self,
        client: str,
        sender_address: str,
        sender_private_key: bytes,
        asset_id: int,
        amount: float,
        fee: float = 1000,
        first_valid_round: int = None,
        last_valid_round: int = None,
        gh: bytes = None,
        transaction_type: str = 'opt-in',
    ):
        self.asset_id = asset_id
        self.amount = amount
        self.transaction_type = transaction_type
        super().__init__(client, sender_address, sender_private_key, fee, first_valid_round, last_valid_round, gh)

    def transact(self):
        tx = AssetTransferTxn(
            sender=self.sender_address,
            fee=self._get_fee(),
            first=self._get_first_valid_round(),
            last=self._get_last_valid_round(),
            gh=self._get_gh(),
            receiver=self.sender_address,
            amt=self.amount,
            index=self.asset_id)

        signed_tx = self.sign_transaction(tx)
        confirm_tx = self.send_transaction(signed_tx)
        return confirm_tx


class AssetTransferTransactor(Transactor):

    def __init__(
        self,
        client: str,
        sender_address: str,
        sender_private_key: bytes,
        asset_id: int,
        amount: float,
        fee: float = 1000,
        first_valid_round: int = None,
        last_valid_round: int = None,
        gh: bytes = None,
        receiver_address: str = None,
        transaction_type: str = 'pay',
    ):
        self.asset_id = asset_id
        self.amount = amount
        self.transaction_type = transaction_type
        self.receiver_address = receiver_address
        super().__init__(client, sender_address, sender_private_key, fee, first_valid_round, last_valid_round, gh)

    def transact(self):
        tx = AssetTransferTxn(
            sender=self.sender_address,
            fee=self._get_fee(),
            first=self._get_first_valid_round(),
            last=self._get_last_valid_round(),
            gh=self._get_gh(),
            receiver=self.receiver_address,
            amt=self.amount,
            index=self.asset_id)

        signed_tx = self.sign_transaction(tx)
        confirm_tx = self.send_transaction(signed_tx)
        return confirm_tx


class SwapAssetTransactor(Transactor):
    def __init__(
        self,
        client: str,
        sender_address: str,
        sender_private_key: bytes,
        asset_id: int,
        asset_amount: float,
        microalgo_amount: float,
        ratn: float,
        ratd: float,
        min_trade: float,
        max_fee: float,
        fee: float = 1000,
        first_valid_round: int = None,
        last_valid_round: int = None,
        gh: bytes = None,
    ):
        self.asset_id = asset_id
        self.asset_amount = asset_amount
        self.microalgo_amount = microalgo_amount
        self.ratn = ratn
        self.ratd = ratd
        self.min_trade = min_trade
        self.max_fee = max_fee
        super().__init__(client, sender_address, sender_private_key, fee, first_valid_round, last_valid_round, gh)

    def transact(self):
        limit = template.LimitOrder(
            owner=self.sender_address,
            asset_id=self.asset_id,
            ratn=self.ratn,
            ratd=self.ratd,
            expiry_round=self.last_valid_round,
            min_trade=self.min_trade,
            max_fee=self.max_fee)
        # addr = limit.get_address()
        program = limit.get_program()
        txns = limit.get_swap_assets_transactions(
            contract=program,
            asset_amount=self.asset_amount,
            microalgo_amount=self.microalgo_amount,
            private_key=self.sender_private_key,
            first_valid=self.first_valid_round,
            last_valid=self.last_valid_round,
            gh=self.gh,
            fee=self.fee)
        # txid = self.client.send_transactions(txns)
        _ = self.client.send_transactions(txns)


class AssetConfigTransactor(Transactor):

    def __init__(
        self,
        client: AlgodClient,
        sender_address: str,
        sender_private_key: bytes,
        total: float,
        unit_name: str,
        asset_name: str,
        fee: float = 1000,
        manager_address: str = None,
        reserve_address: str = None,
        freeze_address: str = None,
        clawback_address: str = None,
        default_frozen: bool = False,
        url: str = None,
        decimal: int = 0,
        note: str = None
    ):
        super().__init__(client, sender_address, sender_private_key)
        self.fee = fee
        self.total = total
        self.unit_name = unit_name
        self.asset_name = asset_name
        self.manager_address = manager_address
        self.reserve_address = reserve_address
        self.freeze_address = freeze_address
        self.clawback_address = clawback_address
        self.default_frozen = default_frozen
        self.url = url
        self.decimal = decimal
        self.note = note

    def transact(self):
        tx = AssetConfigTxn(
            sender=self.sender_address,
            total=self.total,
            fee=self._get_fee(),
            first=self._get_first_valid_round(),
            last=self._get_last_valid_round(),
            gh=self._get_gh(),
            default_frozen=self.default_frozen,
            unit_name=self.unit_name,
            asset_name=self.asset_name,
            manager=self.manager_address,
            reserve=self.reserve_address,
            freeze=self.freeze_address,
            clawback=self.clawback_address,
            url=self.url,
            note=self.note
        )
        signed_tx = self.sign_transaction(tx)
        confirm_tx = self.send_transaction(signed_tx)
        return confirm_tx


class PaymentTransactor(Transactor):

    def __init__(
        self,
        client: AlgodClient,
        sender_private_key: str,
        sender_address: str,
        receiver_address: str,
        amount: float,
        transaction_type: str = 'pay',
        transaction_date: Optional[datetime] = None,
        round_range: int = 1000,
        close_remainder_to_address: str = None,
        note: str = None,
        gen: str = None,
        flat_fee: str = None,
        lease: str = None,
        rekey_to_address: str = None
    ):
        super().__init__(client, sender_address, sender_private_key)
        self.amount = amount
        self.transaction_type = transaction_type
        self.transaction_date = transaction_date
        self.receiver_address = receiver_address
        self.round_range = round_range
        self.close_remainder_to_address = close_remainder_to_address
        self.note = note
        self.gen = gen
        self.flat_fee = flat_fee
        self.lease = lease
        self.rekey_to_address = rekey_to_address

    def transact(self):
        tx = transaction.PaymentTxn(
            sender=self.sender_address,
            fee=self._get_fee(),
            first=self._get_first_valid_round(),
            last=self._get_last_valid_round(),
            gh=self._get_gh(),
            receiver=self.receiver_address,
            amt=self.amount,
            close_remainder_to=self.close_remainder_to_address,
            note=self.note,
            gen=self.gen,
            flat_fee=self.flat_fee,
            lease=self.lease,
            rekey_to=self.rekey_to_address
        )
        signed_tx = self.sign_transaction(tx)
        confirm_tx = self.send_transaction(signed_tx)
        return confirm_tx



import yaml
from pathlib import Path
from pyalgorand.account import Account


class AccountManager:
    def __init__(
        self,
        filename
    ):
        self.filename = filename

    def load_accounts(self):
        if Path(self.filename).exists():
            return yaml.load(open(self.filename))
        else:
            return {}

    def add_account(self, account: Account):
        accounts = self.load_accounts() or {}
        new_account = {
            account.name: {
                'name': account.name,
                'public_address': account.public_address}}
        if hasattr(account, 'nacl_encryption_keys'):
            new_account.update(
                {
                    'nacl_encrytion_public_key': self.nacl_encryption_public_keys.public_key,
                    'nacl_encrytion_private_key': self.nacl_encryption_public_keys._private_key,
                })
        accounts.update(new_account)
        yaml.dump(accounts, open(self.filename, 'w'))



# Create connection
from datetime import datetime, timedelta
from apscheduler.schedulers.background import BackgroundScheduler
from pyalgorand.connector import PureStakeConnector
from pyalgorand.account import Account
from pyalgorand.account_manager import AccountManager
from pyalgorand.transactor import Transactor

credentials = './tmp/.purestake'

con = PureStakeConnector(network='testnet')
con.connect()
client = con.algod_client

# check that the client is working with
client.status()

# Create accounts
alice_account = Account(name='Alice')
alice_account.create()
alice_account.save_pickle(filename='../tmp/alice_account.pkl')

bob_account = Account(name='Bob')
bob_account.create()
bob_account.save_pickle(filename='../tmp/bob_account.pkl')

# once you have created your accounts and saved them, you can just load them
alice_account = Account.from_pickle('../tmp/alice_account.pkl')
bob_account = Account.from_pickle('../tmp/bob_account.pkl')

account_manager = AccountManager('../tmp/account_manager.yml')
account_manager.add_account(alice_account)
account_manager.add_account(bob_account)

# Make transaction
transaction_date = datetime.now() + timedelta(seconds=2)
now = datetime.now()
tor = Transactor(
    client,
    sender_private_key=alice_account.private_key,
    sender_address=alice_account.public_address,
    receiver_address=bob_account.public_address,
    amount=100000,
    transaction_date=transaction_date)

# Start the scheduler
sched = BackgroundScheduler()
sched.start()
# Store the job in a variable in case we want to cancel it
job = sched.add_job(tor.transact, trigger='date', run_date=tor.transaction_date)



import ipfshttpclient
from pathlib import Path
from typing import Union


class IPFS:
    _cached_client = None

    def __init__(
        self,
        url
    ):
        self.url = url
        self.added_on_ipfs = []

    def _connect(self):
        if not self._cached_client:
            self._cached_client = ipfshttpclient.connect(addr=self.url)
        return self._cached_client

    @property
    def client(self):
        return self._connect()

    def add(self, filepath: str, **kwargs):
        item_on_ipfs = self.client.add(filepath, **kwargs)
        self.added_on_ipfs += item_on_ipfs
        return item_on_ipfs

    def delete(self, filepath: str):
        self.client.files.rm(filepath)

    def get_links_from_hash(self, hash):
        oo = self.client.ls(hash)
        return {o['Hash']: o['Links'] for o in oo['Objects']}

    def get(self, hash, local_dir: Union[str, Path] = None, name: str = None):
        self.client.get(hash, local_dir)
        to_file = hash
        if name:
            from_file = local_dir / hash
            to_file = local_dir / name
            Path.rename(from_file, to_file)
        return to_file



from setuptools import find_packages, setup
from pathlib import Path


CURRENT_DIR = Path(__file__).parent

with open(CURRENT_DIR / 'README.md') as readme_file:
    readme = readme_file.read()

# get the dependencies and installs
with open(f'{CURRENT_DIR / "requirements.txt"}', "r", encoding="utf-8") as f:
    # Make sure we strip all comments and options (e.g "--extra-index-url")
    # that arise from a modified pip.conf file that configure global options
    # when running kedro build-reqs
    requires = []
    for line in f:
        req = line.split("#", 1)[0].strip()
        if req and not req.startswith("--"):
            requires.append(req)

setup(
    name='pyalgorand',
    version="0.1.0",
    packages=find_packages(exclude=["tests"]),
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Jean-Baptiste Poullet",
    author_email='jeanbaptistepoullet@gmail.com',
    url='https://bitbucket.org/kuiristo/pyalgorand',
    install_requires=requires,
    license='Apache Software License 2.0',
    extras_require={
        "docs": [
            "sphinx~=3.4.3",
            "sphinx_rtd_theme==0.5.1",
            "nbsphinx==0.8.1",
            "nbstripout==0.3.3",
            "recommonmark==0.7.1",
            "sphinx-autodoc-typehints==1.11.1",
            "sphinx_copybutton==0.3.1",
            "ipykernel~=5.3",
        ]
    }
)



from algosdk import account, mnemonic
import pickle
from typing import ByteString, Optional
from pathlib import Path
from nacl.public import PrivateKey, Box


class Account(object):
    """This class allows to create an Algorand account.
    """
    def __init__(
        self,
        name,
        mode: Optional[str] = 'test'
    ):
        self.name = name
        self.mode = mode

    def create(self):
        private_key, public_address = account.generate_account()
        secret_words = mnemonic.from_private_key(private_key)
        if self.mode == 'test':
            print(
                f'Base64 Private Key: {private_key}\n'
                f'Public Algorand Address: {public_address}\n'
                f'my secret words: {secret_words}'
            )
        self.private_key = private_key
        self.public_address = public_address
        self.secret_words = secret_words

    @staticmethod
    def from_pickle(filename: str or Path):
        return pickle.load(open(filename, 'rb'))

    def save_pickle(self, filename: str or Path):
        pickle.dump(self, open(filename, 'wb'))

    def create_encryption_nacl_keys(self):
        self.nacl_encryption_keys = PrivateKey.generate()
        return self.nacl_encryption_keys

    def encrypt_file_with_nacl(
        self,
        filepath: Path,
        external_public_key: ByteString,
        to_file: Path = None
    ):
        """ Encrypt file with the NaCl encryption system invented by Daniel J. Bernstein.
        """
        # Bob wishes to send Alice an encrypted message so Bob must make a Box with
        #   his private key and Alice's public key
        if not hasattr(self, 'nacl_encryption_keys'):
            self.create_encryption_nacl_keys()
        box = Box(self.nacl_encryption_keys, external_public_key)
        with open(filepath, 'rb') as f:
            encrypted = box.encrypt(f.read())
        if to_file:
            with open(to_file, 'wb') as h:
                h.write(encrypted)
            return
        return encrypted

    def decrypt_file_with_nacl(
        self,
        filepath: Path,
        external_public_key: ByteString
    ):
        """Decrypt file with the NaCl encryption system invented by Daniel J. Bernstein.
        """
        box = Box(self.nacl_encryption_keys, external_public_key)
        with open(filepath, 'rb') as f:
            decrypted = box.decrypt(f.read())
        return decrypted
